#include "Bank.hpp"
#include "Account.hpp"

int main(void)
{
  int account_state;

  //------------------------------------------------------------------------------------------------
  std::cout << "\nÜberprüfungen in Settern / Konstruktoren!" << std::endl;

  Account* my_account = new Account;

  // Bank my_bank(my_account, 0x1337abc); // not valid after check in constructor
  Bank my_bank(my_account, 0x1337); // this is okay

  account_state = my_bank.getAccountEuros(); // pin-code check
  std::cout << "Kontostand: " << account_state << std::endl;


  //------------------------------------------------------------------------------------------------
  std::cout << "\nPointer als Übergabeparameter!" << std::endl;

  my_account->changeEurosBy(-100); // account from parameters, no pin-code check (* in Bank.hpp)

  account_state = my_account->getEuros(); // no pin-code check
  std::cout << "falscher  Kontostand: " << account_state << std::endl;

  account_state = my_bank.getAccountEuros(); // pin-code check
  std::cout << "richtiger Kontostand: " << account_state << std::endl;


  //------------------------------------------------------------------------------------------------
  std::cout << "\nPointer als Rückgabewerte!" << std::endl;

  Account const* my_second_account = my_bank.getAccount(); // getter of account, (# in Bank.hpp)
  // my_second_account->changeEurosBy(-500); // invalid, if pointer is const; no pin-code check

  Account* hacked_account = const_cast<Account*>(my_second_account); // don't use const_cast
  hacked_account->changeEurosBy(-500); // now valid and no pin-code check

  account_state = my_second_account->getEuros(); // no pin-code check
  std::cout << "falscher  Kontostand: " << account_state << std::endl;

  account_state = my_bank.getAccountEuros(); // pin-code check
  std::cout << "richtiger Kontostand: " << account_state << std::endl;

  // Solution: Don't return a reference / pointer, rather try to provide a method
  //           to handle everything inside of the class. If the reference / pointer
  //           has to be returned anyway, make it at least const.


  //------------------------------------------------------------------------------------------------
  std::cout << "\nReinterpret Cast!" << std::endl;
  std::cout << "Größe der Klasse: " << sizeof(my_bank) << std::endl;

  uint8_t* array = reinterpret_cast<uint8_t*>(&my_bank); // interpret the class as a byte-array
  for (size_t i = 0; i < sizeof(my_bank); i += 4)
  {
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i]) << " ";
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i + 1]) << " ";
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i + 2]) << " ";
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i + 3]) << std::endl;
  }
  std::cout << std::endl;

  array[8] = 0x38; // change a byte of the class (no encapsulation)

  for (size_t i = 0; i < sizeof(my_bank); i += 4)
  {
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i]) << " ";
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i + 1]) << " ";
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i + 2]) << " ";
    std::cout.width(2);
    std::cout << std::hex << static_cast<int>(array[i + 3]) << std::endl;
  }
  std::cout << std::endl;

  account_state = my_bank.getAccountEuros();
  std::cout << "richtiger Kontostand: " << std::dec << account_state << std::endl << std::endl;

  // Don't do this in the practicals ever! :D

  return 0;
}
